// src/define-model/transform.ts
import { extractIdentifiers, walkAST } from "ast-walker-scope";

// ../common/src/ast.ts
import { walkIdentifiers } from "@vue/compiler-sfc";
function isCallOf(node, test) {
  return !!(node && node.type === "CallExpression" && node.callee.type === "Identifier" && (typeof test === "string" ? node.callee.name === test : test(node.callee.name)));
}

// ../common/src/constants.ts
var DEFINE_PROPS = "defineProps";
var DEFINE_EMITS = "defineEmits";
var DEFINE_OPTIONS = "defineOptions";
var DEFINE_MODEL = "defineModel";
var REPO_ISSUE_URL = "https://github.com/sxzz/unplugin-vue-macros/issues";

// ../common/src/context.ts
import { MagicString } from "vue/compiler-sfc";

// ../common/src/vue.ts
import { compileScript, parse } from "@vue/compiler-sfc";
var parseSFC = (code, id) => {
  const { descriptor } = parse(code, {
    filename: id
  });
  let compiledScript;
  if (descriptor.script || descriptor.scriptSetup) {
    compiledScript = compileScript(descriptor, {
      id
    });
  }
  return {
    ...descriptor,
    scriptCompiled: compiledScript
  };
};
var addToScript = ({
  sfc: { script, scriptCompiled },
  s,
  scriptCode
}) => {
  if (scriptCode.prepend.length + scriptCode.append.length === 0) {
    return;
  }
  if (script) {
    if (scriptCode.prepend)
      s.appendRight(script.loc.start.offset, scriptCode.prepend);
    if (scriptCode.append)
      s.appendRight(script.loc.end.offset, scriptCode.append);
  } else {
    const lang = scriptCompiled == null ? void 0 : scriptCompiled.attrs.lang;
    const attrs = lang ? ` lang="${lang}"` : "";
    s.prependLeft(
      0,
      `<script${attrs}>${scriptCode.prepend}
${scriptCode.append}<\/script>
`
    );
  }
};

// ../common/src/context.ts
var initContext = (code, id) => {
  let s;
  let sfc;
  const ctx = {
    code,
    id,
    scriptCode: {
      prepend: "",
      append: ""
    },
    get s() {
      return s || (s = new MagicString(code));
    },
    set s(_s) {
      s = _s;
    },
    get sfc() {
      return sfc || (sfc = parseSFC(code, id));
    },
    set sfc(_sfc) {
      sfc = _sfc;
    }
  };
  return {
    ctx,
    getMagicString() {
      return s;
    },
    getSFC() {
      return sfc;
    }
  };
};
var finalizeContext = (ctx) => {
  addToScript(ctx);
};

// ../common/src/unplugin.ts
var getTransformResult = (s, id) => {
  if (s && s.original !== s.toString()) {
    return {
      code: s.toString(),
      get map() {
        return s.generateMap({
          source: id,
          includeContent: true
        });
      }
    };
  }
};

// src/define-model/transform.ts
var transformDefineModel = (ctx, version) => {
  const { s, code } = ctx;
  let hasDefineProps = false;
  let hasDefineEmits = false;
  let hasDefineModel = false;
  let propsTypeDecl;
  let propsDestructureDecl;
  let emitsTypeDecl;
  let emitsIdentifier;
  let modelTypeDecl;
  let modelIdentifier;
  let modelDeclKind;
  let modelDestructureDecl;
  const modelIdentifiers = /* @__PURE__ */ new Set();
  const modelVue2 = { prop: "", event: "" };
  function processDefinePropsOrEmits(node, declId) {
    var _a, _b;
    let type;
    if (isCallOf(node, DEFINE_PROPS)) {
      type = "props";
    } else if (isCallOf(node, DEFINE_EMITS)) {
      type = "emits";
    } else {
      return false;
    }
    const fnName = type === "props" ? DEFINE_PROPS : DEFINE_EMITS;
    if (type === "props")
      hasDefineProps = true;
    else
      hasDefineEmits = true;
    if (node.arguments[0])
      throw new SyntaxError(
        `${fnName}() cannot accept non-type arguments when used with ${DEFINE_MODEL}()`
      );
    const typeDeclRaw = (_b = (_a = node.typeParameters) == null ? void 0 : _a.params) == null ? void 0 : _b[0];
    if (!typeDeclRaw)
      throw new SyntaxError(
        `${fnName}() expected a type parameter when used with ${DEFINE_MODEL}.`
      );
    const typeDecl = resolveQualifiedType(
      typeDeclRaw,
      (node2) => node2.type === "TSTypeLiteral"
    );
    if (!typeDecl) {
      throw new SyntaxError(
        `type argument passed to ${fnName}() must be a literal type, or a reference to an interface or literal type.`
      );
    }
    if (type === "props")
      propsTypeDecl = typeDecl;
    else
      emitsTypeDecl = typeDecl;
    if (declId) {
      if (type === "props" && declId.type === "ObjectPattern") {
        propsDestructureDecl = declId;
      } else if (type === "emits" && declId.type === "Identifier") {
        emitsIdentifier = declId.name;
      }
    } else if (type === "emits") {
      emitsIdentifier = `_${DEFINE_MODEL}_emit`;
      s.prependRight(startOffset + node.start, `const ${emitsIdentifier} = `);
    }
    return true;
  }
  function processDefineModel(node, declId, kind) {
    var _a;
    if (!isCallOf(node, DEFINE_MODEL)) {
      return false;
    }
    if (hasDefineModel) {
      throw new SyntaxError(`duplicate ${DEFINE_MODEL}() call`);
    }
    hasDefineModel = true;
    const propsTypeDeclRaw = (_a = node.typeParameters) == null ? void 0 : _a.params[0];
    if (!propsTypeDeclRaw) {
      throw new SyntaxError(`expected a type parameter for ${DEFINE_MODEL}.`);
    }
    modelTypeDecl = resolveQualifiedType(
      propsTypeDeclRaw,
      (node2) => node2.type === "TSTypeLiteral"
    );
    if (!modelTypeDecl) {
      throw new SyntaxError(
        `type argument passed to ${DEFINE_MODEL}() must be a literal type, or a reference to an interface or literal type.`
      );
    }
    if (declId) {
      const ids = extractIdentifiers(declId);
      ids.forEach((id) => modelIdentifiers.add(id));
      if (declId.type === "ObjectPattern") {
        modelDestructureDecl = declId;
        for (const property of declId.properties) {
          if (property.type === "RestElement") {
            throw new SyntaxError("rest element is not supported");
          }
        }
      } else {
        modelIdentifier = scriptCompiled.loc.source.slice(
          declId.start,
          declId.end
        );
      }
    }
    if (kind)
      modelDeclKind = kind;
    return true;
  }
  function processDefineOptions(node) {
    if (!isCallOf(node, DEFINE_OPTIONS))
      return false;
    const [arg] = node.arguments;
    if (arg)
      processVue2Model(arg);
    return true;
  }
  function processVue2Model(node) {
    if (node.type !== "ObjectExpression")
      return false;
    const model = node.properties.find(
      (prop) => prop.type === "ObjectProperty" && prop.key.type === "Identifier" && prop.key.name === "model" && prop.value.type === "ObjectExpression" && prop.value.properties.length === 2
    );
    if (!model)
      return false;
    model.value.properties.forEach((propertyItem) => {
      if (propertyItem.type === "ObjectProperty" && propertyItem.key.type === "Identifier" && propertyItem.value.type === "StringLiteral" && ["prop", "event"].includes(propertyItem.key.name)) {
        const key = propertyItem.key.name;
        modelVue2[key] = propertyItem.value.value;
      }
    });
    return true;
  }
  function resolveQualifiedType(node, qualifier) {
    if (qualifier(node)) {
      return node;
    }
    if (node.type === "TSTypeReference" && node.typeName.type === "Identifier") {
      const refName = node.typeName.name;
      const isQualifiedType = (node2) => {
        if (node2.type === "TSInterfaceDeclaration" && node2.id.name === refName) {
          return node2.body;
        } else if (node2.type === "TSTypeAliasDeclaration" && node2.id.name === refName && qualifier(node2.typeAnnotation)) {
          return node2.typeAnnotation;
        } else if (node2.type === "ExportNamedDeclaration" && node2.declaration) {
          return isQualifiedType(node2.declaration);
        }
      };
      const body = sfc.scriptCompiled.scriptSetupAst;
      for (const node2 of body) {
        const qualified = isQualifiedType(node2);
        if (qualified) {
          return qualified;
        }
      }
    }
  }
  function extractRuntimeProps(node) {
    const members = node.type === "TSTypeLiteral" ? node.members : node.body;
    const map2 = {};
    for (const m of members) {
      if ((m.type === "TSPropertySignature" || m.type === "TSMethodSignature") && m.key.type === "Identifier") {
        const value = scriptCompiled.loc.source.slice(
          m.typeAnnotation.start,
          m.typeAnnotation.end
        );
        map2[m.key.name] = value;
      }
    }
    return map2;
  }
  function getEventKey(key) {
    if (version === 2) {
      if (modelVue2.prop === key) {
        return modelVue2.event;
      } else if (key === "value") {
        return "input";
      }
    }
    return `update:${key}`;
  }
  function processAssignModelVariable() {
    if (!emitsIdentifier)
      throw new Error(
        `Identifier of returning value of ${DEFINE_EMITS} is not found, please report this issue.
${REPO_ISSUE_URL}`
      );
    const program = {
      type: "Program",
      body: scriptCompiled.scriptSetupAst,
      directives: [],
      sourceType: "module",
      sourceFile: ""
    };
    let hasTransfromed = false;
    walkAST(program, {
      leave(node) {
        if (node.type !== "AssignmentExpression")
          return;
        if (node.left.type !== "Identifier")
          return;
        const idDecl = this.scope[node.left.name];
        if (!modelIdentifiers.has(idDecl))
          return;
        hasTransfromed = true;
        s.overwrite(
          startOffset + node.start,
          startOffset + node.end,
          `__emitHelper(${emitsIdentifier}, '${getEventKey(
            idDecl.name
          )}', ${s.slice(
            startOffset + node.right.start,
            startOffset + node.right.end
          )})`
        );
      }
    });
    if (hasTransfromed) {
      s.prependLeft(
        startOffset,
        "\nimport { emitHelper as __emitHelper } from 'unplugin-vue-macros/helper';"
      );
    }
  }
  if (!code.includes(DEFINE_MODEL))
    return;
  const { sfc } = ctx;
  const { scriptCompiled } = sfc;
  if (!scriptCompiled)
    return;
  const startOffset = scriptCompiled.loc.start.offset;
  if (version === 2 && scriptCompiled.scriptAst && scriptCompiled.scriptAst.length > 0) {
    for (const node of scriptCompiled.scriptAst) {
      if (node.type === "ExportDefaultDeclaration") {
        const { declaration } = node;
        if (declaration.type === "ObjectExpression") {
          processVue2Model(declaration);
        } else if (declaration.type === "CallExpression" && declaration.callee.type === "Identifier" && declaration.callee.name === "defineComponent") {
          declaration.arguments.forEach((arg) => {
            if (arg.type === "ObjectExpression") {
              processVue2Model(arg);
            }
          });
        }
      }
    }
  }
  for (const node of scriptCompiled.scriptSetupAst) {
    if (node.type === "ExpressionStatement") {
      processDefinePropsOrEmits(node.expression);
      if (version === 2) {
        processDefineOptions(node.expression);
      }
      if (processDefineModel(node.expression)) {
        s.remove(node.start + startOffset, node.end + startOffset);
      }
    } else if (node.type === "VariableDeclaration" && !node.declare) {
      const total = node.declarations.length;
      let left = total;
      for (let i = 0; i < total; i++) {
        const decl = node.declarations[i];
        if (decl.init) {
          processDefinePropsOrEmits(decl.init, decl.id);
          if (processDefineModel(decl.init, decl.id, node.kind)) {
            if (left === 1) {
              s.remove(node.start + startOffset, node.end + startOffset);
            } else {
              let start = decl.start + startOffset;
              let end = decl.end + startOffset;
              if (i < total - 1) {
                end = node.declarations[i + 1].start + startOffset;
              } else {
                start = node.declarations[i - 1].end + startOffset;
              }
              s.remove(start, end);
              left--;
            }
          }
        }
      }
    }
  }
  if (!modelTypeDecl)
    return;
  if (modelTypeDecl.type !== "TSTypeLiteral") {
    throw new SyntaxError(
      `type argument passed to ${DEFINE_MODEL}() must be a literal type, or a reference to an interface or literal type.`
    );
  }
  const map = extractRuntimeProps(modelTypeDecl);
  const propsText = Object.entries(map).map(([key, value]) => `${key}${value}`).join("\n");
  const emitsText = Object.entries(map).map(([key, value]) => `(evt: '${getEventKey(key)}', value${value}): void`).join("\n");
  if (hasDefineProps) {
    s.appendLeft(startOffset + propsTypeDecl.start + 1, `${propsText}
`);
    if (propsDestructureDecl && modelDestructureDecl)
      for (const property of modelDestructureDecl.properties) {
        const text = code.slice(
          startOffset + property.start,
          startOffset + property.end
        );
        s.appendLeft(startOffset + propsDestructureDecl.start + 1, `${text}, `);
      }
  } else {
    let text = "";
    const kind = modelDeclKind || "let";
    if (modelIdentifier) {
      text = modelIdentifier;
    } else if (modelDestructureDecl) {
      text = code.slice(
        startOffset + modelDestructureDecl.start,
        startOffset + modelDestructureDecl.end
      );
    }
    s.appendLeft(
      startOffset,
      `
${text ? `${kind} ${text} = ` : ""}defineProps<{
  ${propsText}
}>();`
    );
  }
  if (hasDefineEmits) {
    s.appendLeft(startOffset + emitsTypeDecl.start + 1, `${emitsText}
`);
  } else {
    emitsIdentifier = `_${DEFINE_MODEL}_emit`;
    s.appendLeft(
      startOffset,
      `
const ${emitsIdentifier} = defineEmits<{
  ${emitsText}
}>();`
    );
  }
  if (hasDefineModel) {
    processAssignModelVariable();
  }
  return s;
};

export {
  initContext,
  finalizeContext,
  getTransformResult,
  transformDefineModel
};
